import numpy as np
import matplotlib.pyplot as plt

#=======================================================================

#################
##  FUNCTIONS  ##
#################

###
def produce_generation(prev_pop, prev_fit, mutation_probs, optimal_genotype, n_learn):
	"""
	"""
	#Initialize next generation
	next_pop = []
	next_fit = []
	
	#Create offspring until the next generation is full
	for i in xrange(len(prev_pop)):
		#First pick two parents
		parent_1, parent_2 = sample_parents(prev_pop,prev_fit)
		
		#Then create a new offspring via mating and mutation
		new_offspring = mate_and_mutate(parent_1,parent_2,mutation_probs)
				
		#Add the new offspring to the next generation
		next_pop += [new_offspring]
		
		#Calculate the fitness of this new offspring
		next_fit += [calculate_fitness(new_offspring,optimal_genotype,n_learn)]
		
	#Return the next generation
	return next_pop,next_fit
	
###
def sample_parents(pop,fit):
	"""
	"""
	#Initialize list of parents
	parents = []
	
	#Calculate the max fitness
	max_fit = np.max(fit)
	
	#Find two parents using rejection sampling
	for i in xrange(2):
		while True:
			tmp_ind = np.random.choice(len(pop))
			tmp_rand = np.random.uniform()
			if tmp_rand <= fit[tmp_ind]/max_fit:
				parents += [pop[tmp_ind]]
				break
	
	return parents[0],parents[1]

###
def mate_and_mutate(parent_1,parent_2,mutation_probs):
	"""
	"""
	#Initialize new string
	new_offspring = ''
	
	#Loop over genes
	for i in xrange(len(parent_1)):
		#Pick gene from one of the parents randomly
		tmp_gene = np.random.choice([parent_1[i],parent_2[i]])
		
		#Mutate the gene and store it
		new_offspring += mutation(tmp_gene,mutation_probs)
			
	return new_offspring

###
def mutation(gene,mutation_probs):
	"""
	"""
	return np.random.choice(mutation_probs[gene]['genes'],p=mutation_probs[gene]['probs']) 
	
###
def calculate_fitness(base_genotype,optimal_genotype,n_learn):
	"""
	"""
	#Initialize fitness function
	fitness = float(len(optimal_genotype))
	
	#Loop over learning iterations
	for i in xrange(n_learn + 1):
		if i == 0:
			#Check if correct genotype is known without learning
			if base_genotype == optimal_genotype:
				#No learning is needed, so optimal fitness is achieved
				break
		else:
			#Penalize fitness for required learning
			fitness -= (len(optimal_genotype)-1.)/float(n_learn)
			
			#Do learning
			learned_genotype = learning_trial(base_genotype)
			
			#Check if this round of learning found optimal genotype
			if learned_genotype == optimal_genotype:
				#Learning was successful, optimal fitness is achieved
				break
	
	return fitness

###
def learning_trial(base_genotype):
	"""
	"""
	#Initialized learned genotype
	learned_genotype = ''
	
	#Loop over genotype
	for gene in base_genotype:
		if gene == '?':
			learned_genotype += np.random.choice(['0','1'])
		else:
			learned_genotype += gene

	return learned_genotype

#=======================================================================

###########
##  MAIN ##
###########

def executable(mu, p_opt_geno_evolve, len_genotype, pop_size, n_learn, n_generations, n_sim, outdir):
	"""
	"""
	#Initialize params
	curr_opt_geno = '1'*len_genotype
	
	#Initialize mutation probs
	mutation_probs = {}
	mutation_probs['0'] = {}
	mutation_probs['0']['genes'] = ['0','1','?']
	mutation_probs['0']['probs'] = [1.-mu, 0.5*mu, 0.5*mu]
	mutation_probs['1'] = {}
	mutation_probs['1']['genes'] = ['0','1','?']
	mutation_probs['1']['probs'] = [0.5*mu, 1.-mu, 0.5*mu]
	mutation_probs['?'] = {}
	mutation_probs['?']['genes'] = ['0','1','?']
	mutation_probs['?']['probs'] = [0.5*mu, 0.5*mu, 1.-mu]

	#Initialize optimal genotype evolution probs
	opt_geno_evolve_probs = {}
	opt_geno_evolve_probs['0'] = {}
	opt_geno_evolve_probs['0']['genes'] = ['0','1']
	opt_geno_evolve_probs['0']['probs'] = [1.-p_opt_geno_evolve, p_opt_geno_evolve]
	opt_geno_evolve_probs['1'] = {}
	opt_geno_evolve_probs['1']['genes'] = ['0','1']
	opt_geno_evolve_probs['1']['probs'] = [p_opt_geno_evolve, 1.-p_opt_geno_evolve]

	#Initialize time until 0-fitness
	extinction_time = []
	converge_time = []

	#Loop over simulations
	for sim in xrange(n_sim):
		#Generate initial population
		pop_list = []
		fit_list = []
		plast_list = []
		opt_geno_list = []
		extinction_time += [np.nan]
		converge_time += [np.nan]

		curr_pop = []
		curr_fit = []
		for i in xrange(pop_size):
			#Choose completely random genotype and calculate fitness
			curr_pop += [ ''.join(list(np.random.choice(['0','1','?'],len_genotype))) ]
			curr_fit += [ calculate_fitness(curr_pop[-1],curr_opt_geno,n_learn) ]
		pop_list += [curr_pop]
		fit_list += [curr_fit]
		plast_list += [ np.mean([ np.sum(np.array(list(x)) == '?') for x in curr_pop ]) ]
		opt_geno_list += [curr_opt_geno]

		#Loop over generations
		for i in xrange(n_generations):
			#First evolve the optimal genotype that the child generation fitnesses will be calculated with respect to
			curr_opt_geno = ''
			for gene in opt_geno_list[-1]:
				curr_opt_geno += mutation(gene,opt_geno_evolve_probs)
			opt_geno_list += [ curr_opt_geno ]
			
			#Then mate the parent generation and produce the child generation
			curr_pop, curr_fit = produce_generation(pop_list[-1], fit_list[-1], mutation_probs, opt_geno_list[-1], n_learn)
			pop_list += [curr_pop]
			fit_list += [curr_fit]
			plast_list += [ np.mean([ np.sum(np.array(list(x)) == '?') for x in curr_pop ]) ]
			
			print i, np.mean(fit_list[-1]), np.std(fit_list[-1]), plast_list[-1], opt_geno_list[-1]
			
			#If converged to non-negligible fitness, record it
			if (not converge_time[-1] > -np.inf) and (np.mean(fit_list[-1]) >= 2.):
				converge_time[-1] = i+1
			
			#If extinction happened, break
			if (converge_time[-1] > -np.inf) and np.absolute(np.mean(fit_list[-1]) - 1.0) <= 0.1/pop_size:
				extinction_time[-1] = i+1
				
				print converge_time[-1]
				print extinction_time[-1]
				print '\n'
				
				break

	converge_time = np.array(converge_time)
	extinction_time = np.array(extinction_time)

	print converge_time
	print extinction_time

	#Calculate survival statistics
	frac_converge = np.sum(converge_time >= -np.inf)/float(len(converge_time))
	mean_converge = np.mean( converge_time[converge_time >= -np.inf] )
	std_converge = np.std( converge_time[converge_time >= -np.inf] )

	frac_extinct = np.sum( (extinction_time >= -np.inf)*(converge_time >= -np.inf) )/float(np.sum(converge_time >= -np.inf))
	mean_extinct = np.mean( extinction_time[(extinction_time >= -np.inf)*(converge_time >= -np.inf)] - converge_time[(extinction_time >= -np.inf)*(converge_time >= -np.inf)] )
	std_extinct = np.std( extinction_time[(extinction_time >= -np.inf)*(converge_time >= -np.inf)] - converge_time[(extinction_time >= -np.inf)*(converge_time >= -np.inf)] )	

	print 'Converge: ', frac_converge, mean_converge, std_converge
	print 'Extinct: ', frac_extinct, mean_extinct, std_extinct

	#Save survival statistics
	survive_stats_file = open('%s/survival_statistics_mu_%s_nlearn_%s_ptarget_%s.txt'%(outdir,mu,n_learn,p_opt_geno_evolve),'wt')
	survive_stats_file.write('Config: Genetic Mut. Probability = %s, Learning Rate = %s, Target Mut. Probability = %s\n'%(mu,n_learn,p_opt_geno_evolve))
	survive_stats_file.write('\n')
	survive_stats_file.write('Converged fraction: %s\n'%frac_converge)
	survive_stats_file.write('Converged mean: %s\n'%mean_converge)
	survive_stats_file.write('Converged std: %s\n'%std_converge)
	survive_stats_file.write('\n')
	survive_stats_file.write('Extinction fraction: %s\n'%frac_extinct)
	survive_stats_file.write('Extinction mean: %s\n'%mean_extinct)
	survive_stats_file.write('Extinction std: %s\n'%std_extinct)
	survive_stats_file.close()

	#Plot example timeseries
	myfig=plt.figure()	
	plt.plot(range(len(fit_list)),np.mean(fit_list,axis=-1),'r-',linewidth=3,label='Fitness Mean')
	plt.plot(range(len(fit_list)),np.std(fit_list,axis=-1),'b-',linewidth=3,label='Fitness SD')
	plt.plot(range(len(plast_list)),plast_list,'g-',linewidth=3,label='< N_? >')
	plt.title('Genetic Mut. Probability = %s, Learning Rate = %s, Target Mut. Probability = %s'%(mu,n_learn,p_opt_geno_evolve))
	plt.xlabel('Generation Number')
	plt.legend(loc='best')
	plt.grid(True,which="both")
	plt.xlim(0,1.1*n_generations)
	plt.ylim(0,1.1*len_genotype)
	myfig.savefig('%s/example_timeseries_mu_%s_nlearn_%s_ptarget_%s.png'%(outdir,mu,n_learn,p_opt_geno_evolve), bbox_inches='tight')

	#Run 1000 generations 100 times for each of:
	#P_evolve = 0., 0.001, 0.01
	#n_learn = 2, 10, 100
	#mu = 0., 0.001
	#
	#Can use mean converge times and mean extinction times to find approximation of minimum mutation rate necessary
